import itertools
import shelve
import time
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize
from numba import njit
from numba.typed import List

np.random.seed(0)

###################################### Eikonal Solver ######################################

# |\nabla u| = f
# ((u - a1)^+)^2 + ((u - a2)^+)^2 + ((u - a3)^+)^2 = f^2 h^2


@njit
def calculate_unique_solution(a1, a2, a3, f, h):
    # a1, a2, a3 = np.sort([a1, a2, a3])
    if a1 > a2:
        a1, a2 = a2, a1
    if a2 > a3:
        a2, a3 = a3, a2
    if a1 > a2:
        a1, a2 = a2, a1

    x = a1 + f * h
    if x <= a2:
        return x
    else:
        B = -(a1 + a2)
        C = (a1**2 + a2**2 - f**2 * h**2) / 2
        x1 = (-B + np.sqrt(B**2 - 4 * C)) / 2.0
        x2 = (-B - np.sqrt(B**2 - 4 * C)) / 2.0
        if x1 > a2:
            x = x1
        else:
            x = x2
        if x <= a3:
            return x
        else:
            B = -2.0 * (a1 + a2 + a3) / 3.0
            C = (a1**2 + a2**2 + a3**2 - f**2 * h**2) / 3.0
            x1 = (-B + np.sqrt(B**2 - 4 * C)) / 2.0
            x2 = (-B - np.sqrt(B**2 - 4 * C)) / 2.0
            if x1 > a3:
                x = x1
            else:
                x = x2
        return x


@njit
def sweeping_over_I_J_K(u, I, J, K, f, h):
    m = len(I)
    n = len(J)
    l = len(K)

    for i in I:
        for j in J:
            for k in K:
                if i == 0:
                    uxmin = u[i + 1, j, k]
                elif i == m - 1:
                    uxmin = u[i - 1, j, k]
                else:
                    uxmin = min([u[i - 1, j, k], u[i + 1, j, k]])

                if j == 0:
                    uymin = u[i, j + 1, k]
                elif j == n - 1:
                    uymin = u[i, j - 1, k]
                else:
                    uymin = min([u[i, j - 1, k], u[i, j + 1, k]])

                if k == 0:
                    uzmin = u[i, j, k + 1]
                elif k == l - 1:
                    uzmin = u[i, j, k - 1]
                else:
                    uzmin = min([u[i, j, k - 1], u[i, j, k + 1]])

                u_new = calculate_unique_solution(uxmin, uymin, uzmin, f[i, j, k], h)

                u[i, j, k] = min([u_new, u[i, j, k]])
    return u


@njit
def sweeping(u, v, h):
    f = 1.0 / v  ## slowness

    m, n, l = u.shape
    I = np.arange(m)
    iI = I[::-1]
    J = np.arange(n)
    iJ = J[::-1]
    K = np.arange(l)
    iK = K[::-1]

    u = sweeping_over_I_J_K(u, I, J, K, f, h)
    u = sweeping_over_I_J_K(u, iI, J, K, f, h)
    u = sweeping_over_I_J_K(u, iI, iJ, K, f, h)
    u = sweeping_over_I_J_K(u, I, iJ, K, f, h)

    u = sweeping_over_I_J_K(u, I, iJ, iK, f, h)
    u = sweeping_over_I_J_K(u, I, J, iK, f, h)
    u = sweeping_over_I_J_K(u, iI, J, iK, f, h)
    u = sweeping_over_I_J_K(u, iI, iJ, iK, f, h)

    return u


def eikonal_solve(u, v, h):
    print("Eikonal Solver: ")
    t0 = time.time()
    for i in range(50):
        u_old = np.copy(u)
        u = sweeping(u, v, h)

        err = np.max(np.abs(u - u_old))
        print(f"Iteration {i}, Error = {err}")
        if err < 1e-6:
            break
    print(f"Time: {time.time() - t0:.3f}")
    return u


###################################### Traveltime based on Eikonal Timetable ######################################
@njit
def _get_index(ix, iy, iz, nx, ny, nz, order="C"):
    if order == "C":
        return (ix * ny + iy) * nz + iz
    elif order == "F":
        return (iz * ny + iy) * nx + ix
    else:
        raise ValueError("order must be either C or F")


def test_get_index():
    vx, vy, vz = np.meshgrid(np.arange(3), np.arange(4), np.arange(5), indexing="ij")
    vx = vx.ravel()
    vy = vy.ravel()
    vz = vz.ravel()
    nx = 3
    ny = 4
    nz = 5
    for ix in range(nx):
        for iy in range(ny):
            for iz in range(nz):
                assert vx[_get_index(ix, iy, iz, nx, ny, nz)] == ix
                assert vy[_get_index(ix, iy, iz, nx, ny, nz)] == iy
                assert vz[_get_index(ix, iy, iz, nx, ny, nz)] == iz


@njit
def _interp(time_table, x, y, z, xgrid0, ygrid0, zgrid0, nx, ny, nz, h):
    ix0 = np.floor((x - xgrid0) / h).clip(0, nx - 2).astype(np.int64)
    iy0 = np.floor((y - ygrid0) / h).clip(0, ny - 2).astype(np.int64)
    iz0 = np.floor((z - zgrid0) / h).clip(0, nz - 2).astype(np.int64)
    ix1 = ix0 + 1
    iy1 = iy0 + 1
    iz1 = iz0 + 1

    ## https://en.wikipedia.org/wiki/Bilinear_interpolation
    x0 = ix0 * h + xgrid0
    x1 = ix1 * h + xgrid0
    y0 = iy0 * h + ygrid0
    y1 = iy1 * h + ygrid0
    z0 = iz0 * h + zgrid0
    z1 = iz1 * h + zgrid0

    Q000 = time_table[_get_index(ix0, iy0, iz0, nx, ny, nz)]
    Q001 = time_table[_get_index(ix0, iy0, iz1, nx, ny, nz)]
    Q010 = time_table[_get_index(ix0, iy1, iz0, nx, ny, nz)]
    Q011 = time_table[_get_index(ix0, iy1, iz1, nx, ny, nz)]
    Q100 = time_table[_get_index(ix1, iy0, iz0, nx, ny, nz)]
    Q101 = time_table[_get_index(ix1, iy0, iz1, nx, ny, nz)]
    Q110 = time_table[_get_index(ix1, iy1, iz0, nx, ny, nz)]
    Q111 = time_table[_get_index(ix1, iy1, iz1, nx, ny, nz)]

    t = (
        1.0
        / (x1 - x0)
        / (y1 - y0)
        / (z1 - z0)
        * (
            Q000 * (x1 - x) * (y1 - y) * (z1 - z)
            + Q100 * (x - x0) * (y1 - y) * (z1 - z)
            + Q010 * (x1 - x) * (y - y0) * (z1 - z)
            + Q110 * (x - x0) * (y - y0) * (z1 - z)
            + Q001 * (x1 - x) * (y1 - y) * (z - z0)
            + Q101 * (x - x0) * (y1 - y) * (z - z0)
            + Q011 * (x1 - x) * (y - y0) * (z - z0)
            + Q111 * (x - x0) * (y - y0) * (z - z0)
        )
    )

    return t


# def traveltime(event_loc, station_index, phase_type, eikonal):
def traveltime(event_index, station_index, phase_type, events, stations, eikonal):
    """
    event_index: [N,]
    station_index: [N,]
    phase_type: [N,]
    events: [M, 3]
    stations: [K, 3]
    eikonal: dict
    """
    # x = event_loc[:, 0] - station_loc[:, 0]
    # y = event_loc[:, 1] - station_loc[:, 1]
    # z = event_loc[:, 2] - station_loc[:, 2]
    # x, y, z = event_loc[:, 0], event_loc[:, 1], event_loc[:, 2]
    if isinstance(event_index, int):
        event_index = np.array([event_index] * len(phase_type))
    x, y, z = events[event_index, 0], events[event_index, 1], events[event_index, 2]

    xgrid0 = eikonal["xgrid"][0]
    ygrid0 = eikonal["ygrid"][0]
    zgrid0 = eikonal["zgrid"][0]
    nx = eikonal["nx"]
    ny = eikonal["ny"]
    nz = eikonal["nz"]
    h = eikonal["h"]

    if isinstance(phase_type, list):
        phase_type = np.array(phase_type)
    if isinstance(station_index, list):
        station_index = np.array(station_index)

    tt = np.zeros(len(phase_type), dtype=np.float32)

    p_idx = phase_type == 0
    s_idx = phase_type == 1
    station_unique = np.unique(station_index)
    for i in station_unique:

        sta_idx = station_index == i
        sta_p_idx = sta_idx & p_idx
        sta_s_idx = sta_idx & s_idx

        tt[sta_p_idx] = _interp(
            eikonal["up"][i, :], x[sta_p_idx], y[sta_p_idx], z[sta_p_idx], xgrid0, ygrid0, zgrid0, nx, ny, nz, h
        )
        tt[sta_s_idx] = _interp(
            eikonal["us"][i, :], x[sta_s_idx], y[sta_s_idx], z[sta_s_idx], xgrid0, ygrid0, zgrid0, nx, ny, nz, h
        )

    return tt


def grad_traveltime(event_index, station_index, phase_type, events, stations, eikonal):
    # x = event_loc[:, 0] - station_loc[:, 0]
    # y = event_loc[:, 1] - station_loc[:, 1]
    # z = event_loc[:, 2] - station_loc[:, 2]
    # x, y, z = event_loc[:, 0], event_loc[:, 1], event_loc[:, 2]
    if isinstance(event_index, int):
        event_index = np.array([event_index] * len(phase_type))
    x, y, z = events[event_index, 0], events[event_index, 1], events[event_index, 2]

    xgrid0 = eikonal["xgrid"][0]
    ygrid0 = eikonal["ygrid"][0]
    zgrid0 = eikonal["zgrid"][0]
    nx = eikonal["nx"]
    ny = eikonal["ny"]
    nz = eikonal["nz"]
    h = eikonal["h"]

    if isinstance(phase_type, list):
        phase_type = np.array(phase_type)
    if isinstance(station_index, list):
        station_index = np.array(station_index)

    dt_dx = np.zeros(len(phase_type))
    dt_dy = np.zeros(len(phase_type))
    dt_dz = np.zeros(len(phase_type))

    p_idx = phase_type == 0
    s_idx = phase_type == 1
    station_unique = np.unique(station_index)
    for sta in station_unique:
        sta_idx = station_index == sta
        sta_p_idx = sta_idx & p_idx
        sta_s_idx = sta_idx & s_idx

        dt_dx[sta_p_idx] = _interp(
            eikonal["grad_up"][sta, 0], x[sta_p_idx], y[sta_p_idx], z[sta_p_idx], xgrid0, ygrid0, zgrid0, nx, ny, nz, h
        )
        dt_dx[sta_s_idx] = _interp(
            eikonal["grad_us"][sta, 0], x[sta_s_idx], y[sta_s_idx], z[sta_s_idx], xgrid0, ygrid0, zgrid0, nx, ny, nz, h
        )
        dt_dy[sta_p_idx] = _interp(
            eikonal["grad_up"][sta, 1], x[sta_p_idx], y[sta_p_idx], z[sta_p_idx], xgrid0, ygrid0, zgrid0, nx, ny, nz, h
        )
        dt_dy[sta_s_idx] = _interp(
            eikonal["grad_us"][sta, 1], x[sta_s_idx], y[sta_s_idx], z[sta_s_idx], xgrid0, ygrid0, zgrid0, nx, ny, nz, h
        )
        dt_dz[sta_p_idx] = _interp(
            eikonal["grad_up"][sta, 2], x[sta_p_idx], y[sta_p_idx], z[sta_p_idx], xgrid0, ygrid0, zgrid0, nx, ny, nz, h
        )
        dt_dz[sta_s_idx] = _interp(
            eikonal["grad_us"][sta, 2], x[sta_s_idx], y[sta_s_idx], z[sta_s_idx], xgrid0, ygrid0, zgrid0, nx, ny, nz, h
        )

    grad = np.column_stack((dt_dx, dt_dy, dt_dz))

    return grad


if __name__ == "__main__":

    import os

    data_path = "data"
    if not os.path.exists(data_path):
        os.makedirs(data_path, exist_ok=True)

    # test_get_index()

    nx = 21
    ny = 21
    nz = 21
    vel = {"p": 6.0, "s": 6.0 / 1.73}
    vp = np.ones((nx, ny, nz)) * vel["p"]
    vs = np.ones((nx, ny, nz)) * vel["s"]
    h = 1.0

    sta_ix = 10
    sta_iy = 10
    sta_iz = 10
    station_loc = np.array([sta_ix, sta_iy, sta_iz]) * h
    up = 1000 * np.ones((nx, ny, nz))
    up[sta_ix, sta_iy, sta_iz] = 0.0

    up = eikonal_solve(up, vp, h)
    grad_up = np.gradient(up, h, edge_order=2)
    up = up.ravel()
    grad_up = np.array([x.ravel() for x in grad_up])
    up = up[np.newaxis, :]
    grad_up = grad_up[np.newaxis, :, :]

    us = 1000 * np.ones((nx, ny, nz))
    us[sta_ix, sta_iy, sta_iz] = 0.0

    us = eikonal_solve(us, vs, h)
    grad_us = np.gradient(us, h, edge_order=2)
    us = us.ravel()
    grad_us = np.array([x.ravel() for x in grad_us])
    us = us[np.newaxis, :]
    grad_us = grad_us[np.newaxis, :, :]

    num_event = 10
    event_loc = (
        np.random.rand(num_event, 3) * np.array([nx * h / np.sqrt(2), ny * h / np.sqrt(2), nz * h]) + station_loc
    )
    event_index = np.arange(num_event)
    print(f"{event_loc = }")
    print(f"{station_loc = }")
    # station_loc = np.random.rand(1, 3) * np.array([nx*h/np.sqrt(2), ny*h/np.sqrt(2), 0])
    station_loc = np.tile(station_loc, (num_event, 1))
    station_index = [0] * len(event_loc)
    # phase_type = np.random.choice(["p", "s"], num_event, replace=True)
    phase_type = np.array(["p"] * (num_event // 2) + ["s"] * (num_event - num_event // 2))
    v = np.array([vel[x] for x in phase_type])
    t = np.linalg.norm(event_loc - station_loc, axis=-1, keepdims=False) / v
    grad_t = (
        (event_loc - station_loc) / np.linalg.norm(event_loc - station_loc, axis=-1, keepdims=True) / v[:, np.newaxis]
    )
    print(f"True traveltime: {t = }")
    print(f"True gradient: {grad_t = }")

    config = {
        "up": up,
        "us": us,
        "grad_up": grad_up,
        "grad_us": grad_us,
        "xgrid": np.arange(nx) * h,
        "ygrid": np.arange(ny) * h,
        "zgrid": np.arange(nz) * h,
        "nx": nx,
        "ny": ny,
        "nz": nz,
        "h": h,
    }
    mapping_int = {"p": 0, "s": 1}
    phase_type = np.array([mapping_int[x] for x in phase_type])
    t = traveltime(event_index, station_index, phase_type, event_loc, station_loc, config)
    grad_t = grad_traveltime(event_index, station_index, phase_type, event_loc, station_loc, config)
    print(f"Computed traveltime: {t = }")
    print(f"Computed gradient: {grad_t = }")

    up = np.reshape(up, (nx, ny, nz))
    fig, ax = plt.subplots(1, 3, figsize=(15, 5))
    cax0 = ax[0].pcolormesh(up[sta_ix, :, :])
    fig.colorbar(cax0, ax=ax[0])
    ax[0].invert_yaxis()
    ax[0].set_title("tp_x")
    cax1 = ax[1].pcolormesh(up[:, sta_iy, :])
    fig.colorbar(cax1, ax=ax[1])
    ax[1].invert_yaxis()
    ax[1].set_title("tp_y")
    cax2 = ax[2].pcolormesh(up[:, :, sta_iz])
    fig.colorbar(cax2, ax=ax[2])
    # ax[2].invert_yaxis()
    ax[2].set_title("tp_z")
    plt.savefig(f"{data_path}/slice_tp_3d.png")

    us = np.reshape(us, (nx, ny, nz))
    fig, ax = plt.subplots(1, 3, figsize=(15, 5))
    cax0 = ax[0].pcolormesh(us[sta_ix, :, :])
    fig.colorbar(cax0, ax=ax[0])
    ax[0].invert_yaxis()
    ax[0].set_title("ts_x")
    cax1 = ax[1].pcolormesh(us[:, sta_iy, :])
    fig.colorbar(cax1, ax=ax[1])
    ax[1].invert_yaxis()
    ax[1].set_title("ts_y")
    cax2 = ax[2].pcolormesh(us[:, :, sta_iz])
    fig.colorbar(cax2, ax=ax[2])
    ax[2].invert_yaxis()
    ax[2].set_title("ts_z")
    plt.savefig(f"{data_path}/slice_ts_3d.png")

    grad_up = np.reshape(grad_up, (1, 3, nx, ny, nz))
    fig, ax = plt.subplots(1, 3, figsize=(15, 5))
    cax0 = ax[0].pcolormesh(grad_up[0, 0][sta_ix, :, :])
    fig.colorbar(cax0, ax=ax[0])
    ax[0].invert_yaxis()
    ax[0].set_title("grad_tp_x")
    cax1 = ax[1].pcolormesh(grad_up[0, 1][sta_ix, :, :])
    fig.colorbar(cax1, ax=ax[1])
    ax[1].invert_yaxis()
    ax[1].set_title("grad_tp_y")
    cax2 = ax[2].pcolormesh(grad_up[0, 2][sta_ix, :, :])
    fig.colorbar(cax2, ax=ax[2])
    ax[2].invert_yaxis()
    ax[2].set_title("grad_tp_z")
    plt.savefig(f"{data_path}/slice_grad_tp_3d.png")

    grad_us = np.reshape(grad_us, (1, 3, nx, ny, nz))
    fig, ax = plt.subplots(1, 3, figsize=(15, 5))
    cax0 = ax[0].pcolormesh(grad_us[0, 0][sta_ix, :, :])
    fig.colorbar(cax0, ax=ax[0])
    ax[0].invert_yaxis()
    ax[0].set_title("grad_ts_x")
    cax1 = ax[1].pcolormesh(grad_us[0, 1][sta_ix, :, :])
    fig.colorbar(cax1, ax=ax[1])
    ax[1].invert_yaxis()
    ax[1].set_title("grad_ts_y")
    cax2 = ax[2].pcolormesh(grad_us[0, 2][sta_ix, :, :])
    fig.colorbar(cax2, ax=ax[2])
    ax[2].invert_yaxis()
    ax[2].set_title("grad_ts_z")
    plt.savefig(f"{data_path}/slice_grad_ts_3d.png")



# %%
import torch
import torch.nn as nn
from torch.autograd import Function
from numba import njit
import numpy as np


@njit
def compute(grad_input, grad_output, weight):
    for i in range(grad_output.shape[0]):
        for j in range(grad_output.shape[1]):
            grad_input[i, :] += grad_output[i, j] * weight[j, :]
    return grad_input


# Inherit from Function
class LinearFunction(Function):
    # Note that forward, setup_context, and backward are @staticmethods
    @staticmethod
    def forward(input, weight, bias):
        output = input.mm(weight.t())
        if bias is not None:
            output += bias.unsqueeze(0).expand_as(output)
        return output

    @staticmethod
    # inputs is a Tuple of all of the inputs passed to forward.
    # output is the output of the forward().
    def setup_context(ctx, inputs, output):
        input, weight, bias = inputs
        ctx.save_for_backward(input, weight, bias)

    # This function has only a single output, so it gets only one gradient
    @staticmethod
    def backward(ctx, grad_output):
        # This is a pattern that is very convenient - at the top of backward
        # unpack saved_tensors and initialize all gradients w.r.t. inputs to
        # None. Thanks to the fact that additional trailing Nones are
        # ignored, the return statement is simple even when the function has
        # optional inputs.
        input, weight, bias = ctx.saved_tensors
        grad_input = grad_weight = grad_bias = None

        # These needs_input_grad checks are optional and there only to
        # improve efficiency. If you want to make your code simpler, you can
        # skip them. Returning gradients for inputs that don't require it is
        # not an error.
        if ctx.needs_input_grad[0]:
            grad_input = grad_output.mm(weight)
        if ctx.needs_input_grad[1]:
            grad_weight = grad_output.t().mm(input)
        if bias is not None and ctx.needs_input_grad[2]:
            grad_bias = grad_output.sum(0)

        # print(grad_output.shape, weight.shape)
        grad_output = grad_output.numpy()
        weight = weight.numpy()
        grad_input = np.zeros(input.shape)

        # for i in range(grad_output.shape[0]):
        #     for j in range(grad_output.shape[1]):
        #         grad_input[i, :] += grad_output[i, j] * weight[j, :]
        grad_input = compute(grad_input, grad_output, weight)

        grad_input = torch.from_numpy(grad_input)
        # print(grad_weight)

        return grad_input, grad_weight, grad_bias


class Linear(nn.Module):
    def __init__(self, input_features, output_features, bias=True):
        super().__init__()
        self.input_features = input_features
        self.output_features = output_features

        # nn.Parameter is a special kind of Tensor, that will get
        # automatically registered as Module's parameter once it's assigned
        # as an attribute. Parameters and buffers need to be registered, or
        # they won't appear in .parameters() (doesn't apply to buffers), and
        # won't be converted when e.g. .cuda() is called. You can use
        # .register_buffer() to register buffers.
        # nn.Parameters require gradients by default.
        self.weight = nn.Parameter(torch.empty(output_features, input_features))
        if bias:
            self.bias = nn.Parameter(torch.empty(output_features))
        else:
            # You should always register all possible parameters, but the
            # optional ones can be None if you want.
            self.register_parameter("bias", None)

        # self.weight.requires_grad = False
        # self.bias.requires_grad = False
        # Not a very smart way to initialize weights
        nn.init.uniform_(self.weight, -0.1, 0.1)
        if self.bias is not None:
            nn.init.uniform_(self.bias, -0.1, 0.1)

    def forward(self, input):
        # See the autograd section for explanation of what happens here.
        return LinearFunction.apply(input, self.weight, self.bias)

    def extra_repr(self):
        # (Optional)Set the extra information about this module. You can test
        # it by printing an object of this class.
        return "input_features={}, output_features={}, bias={}".format(
            self.input_features, self.output_features, self.bias is not None
        )


torch.manual_seed(0)
linear = Linear(5, 3)
input = torch.randn(2, 5)
input.requires_grad = True
output = linear(input)
grad_output = output.backward(torch.randn(2, 3))
print(output)
print(grad_output)
# print(linear)

# %%



import multiprocessing as mp
from contextlib import nullcontext

import numpy as np
import pandas as pd
from tqdm import tqdm

from ._ransac import RANSACRegressor


def invert(picks, stations, config, estimator, event_index, event_init):
    """
    Earthquake location for a single event using RANSAC.
    """

    def is_model_valid(estimator, X, y):
        """
        X: idx_sta, type, score, amp
        """
        # n0 = np.sum(X[:, 1] == 0)  # P
        # n1 = np.sum(X[:, 1] == 1)  # S
        n0 = np.sum(X[X[:, 1] == 0, 2])
        n1 = np.sum(X[X[:, 1] == 1, 2])
        if not (n0 >= config["min_p_picks"] and n1 >= config["min_s_picks"]):
            return False

        if estimator.events[0][2] < 1.0:  # depth > 1.0 km
            return False

        if estimator.score(X, y) < config["min_score"]:
            return False

        return True

    def is_data_valid(X, y):
        """
        X: idx_sta, type, score, amp
        y: t_s
        """
        # n0 = np.sum(X[:, 1] == 0)  # P
        # n1 = np.sum(X[:, 1] == 1)  # S
        n0 = np.sum(X[X[:, 1] == 0, 2])
        n1 = np.sum(X[X[:, 1] == 1, 2])
        return n0 >= config["min_p_picks"] and n1 >= config["min_s_picks"]  # At least min P and S picks

    MIN_PICKS = config["min_picks"]
    MIN_PICKS_RATIO = config["min_picks_ratio"]
    if config["use_amplitude"]:
        MAX_RESIDUAL = [config["max_residual_time"], config["max_residual_amplitude"]]
    else:
        MAX_RESIDUAL = config["max_residual_time"]

    if "station_term_time_p" not in stations.columns:
        stations["station_term_time_p"] = 0.0
    if "station_term_time_s" not in stations.columns:
        stations["station_term_time_s"] = 0.0
    if config["use_amplitude"] and ("station_term_amplitude" not in stations.columns):
        stations["station_term_amplitude"] = 0.0

    if config["use_amplitude"]:
        X = picks.merge(
            # stations[["x_km", "y_km", "z_km", "station_id", "station_term_time", "station_term_amplitude"]],
            stations[
                [
                    "x_km",
                    "y_km",
                    "z_km",
                    "station_id",
                    "station_term_time_p",
                    "station_term_time_s",
                    "station_term_amplitude",
                ]
            ],  ## Separate P and S station term
            on="station_id",
        )
    else:
        X = picks.merge(
            # stations[["x_km", "y_km", "z_km", "station_id", "station_term_time"]],
            stations[
                ["x_km", "y_km", "z_km", "station_id", "station_term_time_p", "station_term_time_s"]
            ],  ## Separate P and S station term
            on="station_id",
        )
    t0 = X["phase_time"].min()

    if (event_init is None) or len(X) == 0:
        event_init = np.array([[np.median(X["x_km"]), np.median(X["y_km"]), 5.0, 0.0]])
    else:
        event_init = np.array([[event_init[0], event_init[1], event_init[2], (event_init[3] - t0).total_seconds()]])

    # xstd = np.std(X["x_km"])
    # ystd = np.std(X["y_km"])
    # rstd = np.sqrt(xstd**2 + ystd**2)

    X.rename(columns={"phase_type": "type", "phase_score": "score", "phase_time": "t_s"}, inplace=True)
    X["t_s"] = (X["t_s"] - t0).dt.total_seconds()
    X["station_term_time"] = X.apply(
        lambda x: x["station_term_time_p"] if x["type"] == 0 else x["station_term_time_s"], axis=1
    )  ## Separate P and S station term
    X["t_s"] = X["t_s"] - X["station_term_time"]
    if config["use_amplitude"]:
        X.rename(columns={"phase_amplitude": "amp"}, inplace=True)
        X["amp"] = X["amp"] - X["station_term_amplitude"]

    estimator.set_params(**{"events": event_init})

    # ## Location using ADLoc
    # if config["use_amplitude"]:
    #     estimator.fit(X[["idx_sta", "type", "score", "amp"]].values, y=X[["t_s", "amp"]].values)
    # else:
    #     estimator.fit(X[["idx_sta", "type", "score"]].values, y=X[["t_s"]].values)
    # inlier_mask = np.ones(len(X)).astype(bool)

    ## Location using RANSAC
    num_picks = len(X)
    reg = RANSACRegressor(
        estimator=estimator,
        random_state=0,
        min_samples=max(MIN_PICKS, int(MIN_PICKS_RATIO * num_picks)),
        # residual_threshold=MAX_RESIDUAL * (1.0 - np.exp(-rstd / 60.0)),  # not sure which one is better
        residual_threshold=MAX_RESIDUAL,
        is_model_valid=is_model_valid,
        is_data_valid=is_data_valid,
    )
    try:
        if config["use_amplitude"]:
            reg.fit(X[["idx_sta", "type", "score", "amp"]].values, X[["t_s", "amp"]].values)
        else:
            reg.fit(X[["idx_sta", "type", "score"]].values, X[["t_s"]].values)
    except Exception as e:
        # print(f"No valid model for event_index {event_index}.")
        message = "RANSAC could not find a valid consensus set."
        if str(e)[: len(message)] != message:
            print(e)
            raise
        picks["mask"] = 0
        picks["residual_time"] = 0.0
        if config["use_amplitude"]:
            picks["residual_amplitude"] = 0.0
        return picks, None

    estimator = reg.estimator_
    inlier_mask = reg.inlier_mask_

    ## Predict travel time
    output = estimator.predict(X[["idx_sta", "type"]].values)
    if config["use_amplitude"]:
        tt = output[:, 0]
        amp = output[:, 1]
    else:
        tt = output[:, 0]
    if config["use_amplitude"]:
        score = estimator.score(
            X[["idx_sta", "type", "score", "amp"]].values[inlier_mask], y=X[["t_s", "amp"]].values[inlier_mask]
        )
    else:
        score = estimator.score(X[["idx_sta", "type", "score"]].values[inlier_mask], y=X[["t_s"]].values[inlier_mask])

    if (np.sum(inlier_mask) > MIN_PICKS) and (score > config["min_score"]):
        mean_residual_time = np.sum(np.abs(X["t_s"].values - tt) * inlier_mask) / np.sum(inlier_mask)
        if config["use_amplitude"]:
            mean_residual_amp = np.sum(np.abs(X["amp"].values - amp) * inlier_mask) / np.sum(inlier_mask)
        else:
            mean_residual_amp = 0.0
        x, y, z, t = estimator.events[0]
        if config["use_amplitude"]:
            mag = estimator.magnitudes[0]

        event = {
            "idx_eve": event_index,  ## inside adloc, idx_eve is used which starts from 0 to N events
            "x_km": x,
            "y_km": y,
            "z_km": z,
            "time": t0 + pd.Timedelta(t, unit="s"),
            "adloc_score": score,
            "adloc_residual_time": mean_residual_time,
            "num_picks": np.sum(inlier_mask),
        }
        if config["use_amplitude"]:
            event["magnitude"] = mag
            event["adloc_residual_amplitude"] = mean_residual_amp
        event = pd.DataFrame([event])
    else:
        inlier_mask = np.zeros(len(inlier_mask), dtype=bool)
        event = None

    picks["residual_time"] = X["t_s"].values - tt
    if config["use_amplitude"]:
        picks["residual_amplitude"] = X["amp"].values - amp
    else:
        picks["residual_amplitude"] = 0.0
    picks["mask"] = inlier_mask.astype(int)

    # ####################################### Debug #######################################
    # import os

    # import matplotlib.pyplot as plt

    # print(f"{max(MIN_PICKS, int(MIN_PICKS_RATIO * num_picks))=}")
    # print(f"{MAX_RESIDUAL=}")
    # fig, ax = plt.subplots(1, 2, squeeze=False, figsize=(7, 4))
    # event = estimator.events[0]
    # X["dist"] = np.linalg.norm(X[["x_km", "y_km", "z_km"]].values - event[:3], axis=1)
    # ax[0, 0].scatter(X["dist"], X["t_s"], c="r", s=10, label="Picks")
    # ax[0, 0].scatter(X["dist"][~inlier_mask], X["t_s"][~inlier_mask], s=10, c="k", label="Noise")
    # ax[0, 0].scatter(X["dist"][inlier_mask], tt[inlier_mask], c="g", s=30, marker="x", label="Predicted")
    # ax[0, 0].set_xlabel("Distance (km)")
    # ax[0, 0].set_ylabel("Time (s)")
    # ax[0, 0].legend()
    # ax[0, 1].scatter(np.log10(X["dist"]), X["amp"], c="r", s=10, label="Picks")
    # ax[0, 1].scatter(np.log10(X["dist"][~inlier_mask]), X["amp"][~inlier_mask], s=10, c="k", label="Noise")
    # ax[0, 1].scatter(np.log10(X["dist"][inlier_mask]), amp[inlier_mask], c="g", s=30, marker="x", label="Predicted")
    # ax[0, 1].set_xlabel("Distance (km)")
    # ax[0, 1].set_ylabel("Amplitude")
    # if not os.path.exists("debug"):
    #     os.makedirs("debug")
    # fig.savefig(f"debug/event_{event_index}.png")
    # plt.close(fig)
    # raise
    # ####################################### Debug #######################################

    return picks, event


def invert_location(picks, stations, config, estimator, events_init=None, iter=0):

    if "ncpu" in config:
        NCPU = config["ncpu"]
    else:
        NCPU = min(64, mp.cpu_count() * 2 - 1)

    jobs = []
    num_events = len(picks["idx_eve"].unique())
    events_inverted = []
    picks_inverted = []
    pbar = tqdm(total=num_events, desc=f"Iter {iter}")
    with mp.get_context("spawn").Pool(NCPU) as pool:
        for idx_eve, picks_by_event in picks.groupby("idx_eve"):
            if events_init is not None:
                event_init = events_init[events_init["idx_eve"] == idx_eve]
                if len(event_init) == 0:
                    event_init = None
                elif len(event_init) == 1:
                    event_init = event_init[["x_km", "y_km", "z_km", "time"]].values[0]
                else:
                    event_init = event_init[["x_km", "y_km", "z_km", "time"]].values[0]
                    print(f"Multiple initial locations for event_index {idx_eve}.")
            else:
                event_init = None

            # picks_, event_ = invert(picks_by_event, stations, config, estimator, idx_eve, event_init)
            # if event_ is not None:
            #     events_inverted.append(event_)
            # picks_inverted.append(picks_)
            # pbar.update()

            job = pool.apply_async(
                invert,
                (picks_by_event, stations, config, estimator, idx_eve, event_init),
                callback=lambda _: pbar.update(),
            )
            jobs.append(job)

        pool.close()
        pool.join()

        for job in jobs:
            picks_, event_ = job.get()
            if event_ is not None:
                events_inverted.append(event_)
            picks_inverted.append(picks_)

    if len(events_inverted) == 0:
        return None, None

    events_inverted = pd.concat(events_inverted, ignore_index=True)
    picks_inverted = pd.concat(picks_inverted)
    if events_init is not None:
        events_inverted = events_inverted.merge(events_init[["event_index", "idx_eve"]], on="idx_eve")

    print(f"ADLoc locates {len(events_inverted)} events outof {num_events} events")
    print(f"using {len(picks_inverted[picks_inverted['mask'] == 1])} picks outof {len(picks_inverted)} picks")

    return picks_inverted, events_inverted


def invert_location_iter(picks, stations, config, estimator, events_init=None, iter=0):

    events_inverted = []
    picks_inverted = []

    for idx_eve, picks_by_event in tqdm(picks.groupby("idx_eve"), desc=f"Iter {iter}"):
        if events_init is not None:
            event_init = events_init[events_init["idx_eve"] == idx_eve]
            if len(event_init) > 0:
                event_init = event_init[["x_km", "y_km", "z_km", "time"]].values[0]
        else:
            event_init = None

        picks_, event_ = invert(picks_by_event, stations, config, estimator, idx_eve, event_init)

        if event_ is not None:
            events_inverted.append(event_)
        picks_inverted.append(picks_)

    if len(events_inverted) == 0:
        return None, None
    events_inverted = pd.concat(events_inverted, ignore_index=True)
    picks_inverted = pd.concat(picks_inverted)
    if events_init is not None:
        events_inverted = events_inverted.merge(events_init[["event_index", "idx_eve"]], on="idx_eve")
    else:
        events_inverted["event_index"] = events_inverted["idx_eve"]

    print(f"ADLoc using {len(picks_inverted[picks_inverted['mask'] == 1])} picks outof {len(picks_inverted)} picks")
    return picks_inverted, events_inverted
